#!/usr/bin/env python

# Top level script for invoking Wing.  If --use-src is specified
#  as an arg, then the files in WINGHOME/src, WINGHOME/external,
#  WINGHOME/opensource will be used; otherwise, the files in the version
#  specific bin directory will be used if it exists.
#
# Copyright (c) 2000-2012, Archaeopteryx Software, Inc.  All rights reserved.
#
# Written by Stephan R.A. Deibel and John P. Ehresman
import sys
import os

# XXX Temporary? flag for showing gc debug output
if '--gc-debug-stats' in sys.argv:
  sys.argv.remove('--gc-debug-stats')
  import gc
  gc.set_debug(gc.DEBUG_STATS)
  
if '--verbose-trace' in sys.argv:
  def trace(frame, event, arg):
    try:
      fname = repr(frame.f_code)
    except:
      fname = '<unknown>'
    sys.__stderr__.write(event + ' ' + fname) 
    sys.__stderr__.write('\n')
  sys.setprofile(trace)

# Hack to set default encoding to utf-8 because most Wing code was
# developed to run with pygtk, which also changed the default encoding
# Since Wing changes the default encoding, there are potential bugs
# with any externally developed module that assumes a different default
# encoding.  Remember that dealing with encodings is _hard_ even though
# the most developers assume is that it should be easy (rather like
# threads or floating point).  The encoding story changes with Python 3.
def SetDefaultEncodingToUtf8():

  # XXXQT May want to change this to calling PyUnicode_SetDefaultEncoding
  # XXXQT directly and avoid reloading sys, but this works for now.
  
  try:
    sys.setdefaultencoding
  except AttributeError:
    # Note that reloading resets sys.excepthook, which needs to be set for
    # exception reporting to work in the debugger, and stdin / stdout / stderr,
    # which are needed for Wing's test runner, so the values are saved &
    # then reset
    to_save_names = ['excepthook', 'stdin', 'stdout', 'stderr']
    saved = dict((name, getattr(sys, name)) for name in to_save_names)
    reload(sys)
    for name, val in saved.iteritems():
      setattr(sys, name, val)

  sys.setdefaultencoding('utf-8')

SetDefaultEncodingToUtf8()
  
PATCHES_INSTALLED = False
kWingHome = None

def RelativePathJoin(*parts):
  """ Join where 2nd+ parts are relative. """
  
  if __debug__:
    for p in parts[1:]:
      assert not os.path.isabs(p)

  return os.path.join(*parts)


def FindWingHome(argv):
  """ Returns abspath of winghome as an unicode object. """
  
  fs_encoding = sys.getfilesystemencoding()
  if fs_encoding is None:
    fs_encoding = sys.getdefaultencoding()
  if '--utf8-args' in sys.argv:
    encoding = 'utf_8'
  else:
    encoding = fs_encoding
  
  winghome = None
  
  # This lets us debug the IDE with an installed copy of itself on win32
  if '--use-devel-winghome' in sys.argv:
    i = argv.index('--use-devel-winghome')
    winghome = argv[i + 1]
    del argv[i:i + 2]
    
  # This is used when running the binary installation on win32
  elif '--use-winghome' in sys.argv:
    i = argv.index('--use-winghome')
    winghome=argv[i + 1]
    del argv[i:i + 2]
    
  # Look in environment for a given value (used on Posix and on win32
  # when running Wing from sources)
  elif os.environ.has_key('WINGHOME'):
    winghome=os.environ['WINGHOME']
    encoding = fs_encoding
  
  # If none given, guess WINGHOME based on location of this script
  else:
    winghome = os.path.dirname(os.path.dirname(os.path.abspath(sys.argv[0])))
  
  # Hack so I can debug this with Wing 1.1.6 [SRAD]
  if os.environ.has_key('FORCE_WINGHOME'):
    winghome = os.environ['FORCE_WINGHOME']
    encoding = fs_encoding

  if winghome is None:
    return None
  
  if encoding is None:
    encoding = sys.getdefaultencoding()
  winghome = unicode(winghome, encoding, 'strict')

  if not os.path.isabs(winghome):
    try:
      cwd = os.getcwdu()
    except Exception:
      pass
    else:
      winghome = os.path.join(cwd, winghome)
      
  return os.path.normpath(winghome)


def FindBasePath(log_func):
  """ Find the base directory to load .py / .pyc files from.  Throws System exception
  if no base path can be found """

  # Caller requested forcing execution from source code even if binaries are available
  if '--use-src' in sys.argv:
    base_path = kWingHome
    log_func("Running Wing from source code in %s\n" % base_path)
    sys.argv.remove('--use-src')
    return base_path

  # Otherwise locate the wing executable base path:  We look through bin dir
  # in sub-directory according to the Python major.minor version that's running, and
  # try all lesser releases with same major.minor number if we don't
  # have an exact match
  major, minor, release = sys.version_info[:3]
  log_func("Running Wing with Python version %d.%d.%d\n" % (major, minor, release))
  success = 0
  version_id = '%d.%d' % (major, minor)
  
  # Binaries in bin/ide-2.7
  base_path = RelativePathJoin(kWingHome, 'bin', 'ide-' + version_id)
  if os.path.isdir(base_path):
    log_func("Using precompiled binaries for Python version " + version_id + "\n")
    return base_path

  # Binaries in bin/2.7
  base_path = RelativePathJoin(kWingHome, 'bin', version_id)
  if os.path.isdir(base_path):
    log_func("Using precompiled binaries for Python version " + version_id + "\n")
    return base_path

  # If not found, then try to run from source code
  base_path = kWingHome
  # Check whether we managed to find a valid base path
  if not os.path.isdir(base_path):
    log_func("Running with invalid WINGHOME: %s\n" % base_path)
    log_func("Please check your installation\n")
    sys.exit(-1)

  log_func("Running from source code files.\n")

  # Run with pychecker if it's present
  pychecker_dir = os.path.normpath(RelativePathJoin(kWingHome, '..', 'pychecker'))
  if os.path.isdir(pychecker_dir):
    os.environ['PYCHECKER'] = 'errors no-imports no-pkgimport no-local no-argsused no-varargsused no-unusednames'
    sys.path.append(pychecker_dir)
    import pychecker.checker
      
  return base_path

def _GetQtRuntimeVersion():
  """ Gets the name of the qt runtime in bin as 'qt#.#'. """

  try:
    name_list = os.listdir(RelativePathJoin(kWingHome, 'bin'))
  except OSError:
    name_list = []
    
  for name in name_list:
    if name.startswith('runtime-qt'):
      return name[len('runtime-'):]
    
  return 'qt'  # Probably wrong
  

def _GetInternalPaths():
  """ Return list of relative pathnames """
    
  qt_rt = _GetQtRuntimeVersion()
  
  runtimes_with_python = [
    'sip-py2.7',
    'pyqt5-py2.7-' + qt_rt,
    'scintillaedit-' + qt_rt,
  ]
  internal_path_list = [
    'src',
    'opensource',
    'external/docutils/extras',
    'external/docutils',
    'external/faulthandler',
    'external/Imaging-1.1.7',
  ]
  
  lib_subdir = ('lib/site-packages' if sys.platform == 'win32'
                 else 'lib/python2.7/site-packages')
  
  internal_path_list.extend(['bin/runtime-%s/%s' % (n, lib_subdir)
                             for n in runtimes_with_python])

  if __debug__:
    internal_path_list.append('build-files/build-temp/cython')
  internal_path_list.append('external/pyside-lgpl')
  
  return internal_path_list

def SetupPythonPath(log_func):
  """ Adds directories to python path. """

  base_path = FindBasePath(log_func)

  internal_paths = _GetInternalPaths()
  for path in internal_paths:
    parts = path.split('/')
    if parts[0] == 'bin':
      full_name = RelativePathJoin(kWingHome, *parts)
    else:
      zip_name = RelativePathJoin(base_path, parts[0] + '.zip')
      if os.path.exists(zip_name):
        full_name = RelativePathJoin(zip_name, *parts[1:])
      else:
        full_name = RelativePathJoin(base_path, *parts)
        
    if sys.path[0] != full_name:
      if full_name in sys.path:
        sys.path.remove(full_name)
      sys.path.insert(0, full_name)

def GetSquelchOutputFlag():
  if '--squelch-output' in sys.argv:
    sys.argv.remove('--squelch-output')
    return True
  else:
    return False
  

def GetProfileFlag():
  # Override profile setting is --profile is in argv
  if '--profile' in sys.argv:
    sys.argv.remove('--profile')
    return True
  else:
    return False
  
def GetDebugFlag():
  """ Check whether issued debug option which turns off blanket 
  exception handlers so that the Wing debugger can report 
  exceptions right away. """
  if '--debug' in sys.argv:
    sys.argv.remove('--debug')
    return True
  else:
    return __debug__

#########################################################################
# Add dynamic library and patch directories
#########################################################################

def InitializeDynamicPaths(log_func):
  """ Init dynamic (non-standard paths), including patch directories.  Note
  the following are imported before patch dirs are added: config (w/ win32misc on
  win32), vinfo, capabilities, encoding_utils, & mergeimporter. """
  
  from wingbase import mergeimporter
  meta_path = mergeimporter.MergeDirImporter()
  sys.meta_path.insert(0, meta_path)

  _AddDynamicLibDirs(meta_path, log_func)
  _AddPatchDirs(meta_path, log_func)

def _AddDynamicLibDirs(importer, log_func):
  """Add dynamic library directories -- needed when the pyo/pyc files are stored
  in a zip archive, since dynamic libs cannot be loaded from zips due to OS
  limitations"""
  
  internal_paths = _GetInternalPaths()
  base_path = FindBasePath(log_func)
  full_paths = [RelativePathJoin(base_path, *dir.split('/'))
                for dir in internal_paths]
  for full in full_paths:
    full = os.path.normpath(full)
    full_parts = full.split(os.sep)
    if os.path.isdir(full):
      importer.add_dir(full)
      
      # Add a cython-compiled subdir as if the modules were in the package
      try:
        descendant_dirs = [extended for extended, s1, s2 in os.walk(full)
                           if extended != full]
      except:
        descendant_dirs = []
        
      for subdir in descendant_dirs:
        subdir = os.path.normpath(subdir)
        subdir_parts = subdir.split(os.sep)
        if subdir_parts[-1] == 'cython-compiled':
          pkg_name_parts = subdir_parts[len(full_parts):-1]
          if len(pkg_name_parts) != 0:
            pkg_prefix = '.'.join(pkg_name_parts) + '.'
            importer.add_dir(subdir, pkg_prefix)

  log_func("DYNLIB_DIRS=%s\n" % ', '.join(full_paths))

def _GetPatchDirs(wing_version, winghome, userwingdir):
  """ Get (name, dirname) pairs for all patches. Use all dirs from both wing
  home & user settings dir but reverse sort by local name so names with >
  alphabetical order take precedence. Patches are preceded by the patch number
  in ### format, so this lets a newer patch override files in an older one.
  Returned list is a new list instance and is sorted in alpha order. """
  
  dirs = [
    RelativePathJoin(winghome, 'patches', wing_version),
    RelativePathJoin(userwingdir, 'patches', wing_version)
  ]

  pair_list = []
  for patch_dir in dirs:
    try:
      name_list = os.listdir(patch_dir)
    except (IOError, OSError):
      name_list = []
    else:
      pair_list.extend([(name, RelativePathJoin(patch_dir, name))
                        for name in name_list])
  pair_list.sort()
  return pair_list  

def _AddPatchDirs(importer, log_func):
  """Add importable files in patch directories. """

  global PATCHES_INSTALLED
  assert not PATCHES_INSTALLED

  import minimalconfig
  
  del minimalconfig.gPatchPairList[:]
  # Don't load patches when working from sources
  if __debug__:
    PATCHES_INSTALLED = True
    return
  
  wing_version = minimalconfig.kVersion
  if minimalconfig.kBuild.startswith('b') or minimalconfig.kBuild.startswith('rc'):
    wing_version += minimalconfig.kBuild

  internal_path_seq = _GetInternalPaths()

  pair_list = _GetPatchDirs(wing_version, kWingHome, minimalconfig.kUserWingDir)
  for name, dirname in pair_list:
    for internal_path in internal_path_seq:
      if internal_path.startswith('bin/'):
        possible_dir = RelativePathJoin(dirname, internal_path)
      else:
        possible_dir = RelativePathJoin(dirname, 'bin', 'ide-2.7', internal_path)
        
      if os.path.isdir(possible_dir):
        log_func("Adding patch dir to merge importer: {}\n".format(possible_dir))
        importer.add_dir(possible_dir)
      
  minimalconfig.gPatchPairList.extend(pair_list)

  PATCHES_INSTALLED = True

  log_func("PATCHES=%s\n" % ', '.join([p[0] for p in pair_list]))

def InitializeDebugServer():

  assert PATCHES_INSTALLED

  from debug.tserver import abstract
  abstract._SetWingHome(kWingHome)

###########################################################################
# Save startup environment variables in config.gStartupEnv. This is used
# later when spawning sub-processes to avoid polluting them with environment
# values we set inside Wing's startup scripts (see runwing.sh). The original
# values are obtained from the ORIG_ settings, if the ORIG_ variables are
# present. If the ORIG_ variables are set to __undefined__, then we del the
# associated env variable from the env.
###########################################################################
  
def _GetStartupEnv():
  from wingbase import miscutils
  env = miscutils.GetUnicodeEnv()
  for key, orig_val in env.items():
    
    # Found an overridden key
    if key.startswith('WING_ORIG_'):
      
      # Get the original value and environment variable name
      orig_key = key[len('WING_ORIG_'):]
  
      # If the environment had a valid original value, then restore that
      if orig_val != '__undefined__':
        env[orig_key] = orig_val
      
      # If we set the environment variable but it wasn't originally set, then
      # just remove our temporary value
      else:
        try:
          del env[orig_key]
        except KeyError:
          pass
        
      # Also delete our override key, since it's no longer needed
      del env[key]
 
  # Strip out any non-string keys or values
  for key, val in list(env.iteritems()):
    if not isinstance(key, basestring) or not isinstance(val, basestring):
      try:
        del env[key]
      except KeyError:
        pass

  return env      
      
def InitializeConfig():

  assert PATCHES_INSTALLED
  
  import config
  config.gStartupEnv = _GetStartupEnv()


def InitializeStdoutLogging(squelch_output):
  # Redirect stdout and stderr to our logging implementation
  # This import will fail if something is wrong with the
  # installation so we make sure we output errors when it does
  
  assert PATCHES_INSTALLED

  import config

  try:
    import sys
    import wingutils.reflect
  except ImportError:
    sys.exit(1)
    
  sys.stderr = wingutils.reflect.CLoggedOutput()
  sys.stdout = wingutils.reflect.CLoggedOutput()
  
  # Set up exception log
  if squelch_output:
    logs = [ config.kErrorLogFile ]
  else:  
    logs = [ sys.__stderr__, config.kErrorLogFile ]
  threshold = None  # Don't truncate during startup
  wingutils.reflect.ConfigureExceptionReporting(logs, threshold)
  
def InitializeFaultHandler(log, dirname):

  # Don't do this on Windows because we already use win32misc.InWin32TryExcept
  # to log fatal errors there
  if sys.platform == 'win32':
    return
  
  import traceback
  import time
  import stat
  import faulthandler
  
  segfault_fn = os.path.join(dirname, 'segfault.log')

  try:
    if os.path.exists(segfault_fn) and os.stat(segfault_fn)[stat.ST_SIZE] > 1024 * 1024:
      os.unlink(segfault_fn)
  except:
    log("Failed to inspect or remove segfault log %s\n", segfault_fn)
    log(traceback.format_exc())
    
  try:
    f = open(segfault_fn, 'a')
    f.write(time.strftime('wingide - %Y/%m/%d - %H:%M:%S - Startup\n'))
    f.flush()
    faulthandler.enable(f)
    # IMPORTANT:  DO NOT CLOSE THE FILE!
    # We leave the file open for the life of the process because
    # faulthandler uses the file descriptor and not the open file
    # object.  If we close the file the file description could be 
    # reused by another file and the stack data would be written
    # there instead.
  except:
    log("Failed to set up segfault log %s\n" % segfault_fn)
    log(traceback.format_exc())
  else:
    log("Set up segfault log %s\n" % segfault_fn)
    
def InitializeSteam(log, winghome):
  """ Initializes Steam library, if present """

  import platform
  if sys.platform == 'win32':

    from wingbase import win32
    
    steam_lib = os.path.normpath(winghome + '/shared/steam/lib/win32/steam_api.dll')
    if not os.path.exists(steam_lib):
      return
    
    try:
      log("Loading %s\n" % steam_lib)
      win32.LoadLibrary(steam_lib)
    except WindowsError:
      log('Unable to load %s\n' % steam_lib)
      raise
    
  elif sys.platform.startswith('linux'):
    
    import ctypes
    
    if sys.maxint > 2 ** 32:
      steam_lib = os.path.normpath(winghome + '/../shared/steam/lib/linux64/libsteam_api.so')
    else:
      steam_lib = os.path.normpath(winghome + '/../shared/steam/lib/linux32/libsteam_api.so')
      
    log('STEAMLIB=%s\n'%steam_lib)
    if not os.path.exists(steam_lib):
      return
    
    try:
      log("Loading %s\n" % steam_lib)
      ctypes.cdll.LoadLibrary(steam_lib)
    except:
      log('Unable to load %s\n' % steam_lib)
      raise
    
    log("Loaded %s\n" % steam_lib)
    
  elif sys.platform == 'darwin':
    # OS X setup is in run-wing.sh using DYLD_LIBRARY_PATH
    pass
  
def InitializeQt(log, winghome):
  """ Initializes environment for Qt """
  
  import platform
  
  qt_runtime = _GetQtRuntimeVersion()
  
  assert 'guiutils.qt_config' not in sys.modules
  from guiutils import qt_config
  qt_config.kUsePySide = False
  use_webkit = False
  if '--enable-webkit' in sys.argv:
    sys.argv.remove('--enable-webkit')
    use_webkit = True
  if '--no-webkit' in sys.argv:
    sys.argv.remove('--no-webkit')
    use_webkit = False
  if qt_runtime != 'qt5.5':
    use_webkit = False
    
  if not use_webkit:
    qt_config.kUseWebKit = False
    qt_config.kUseTextBrowser = True

  InitializePyQt(log, winghome, qt_runtime)
  
  # Load wgtk after Qt libs are loaded
  from guiutils import wgtk

  InitializePalettes()
  
  # Work around qt hardcoding the OS X system font and it changing w/ OS X 10.9
  if sys.platform == 'darwin':
    release, versioninfo, machine = platform.mac_ver()
    if release >= '10.9':
      wgtk.QFont.insertSubstitution(".Lucida Grande UI", "Lucida Grande")
  
  qt_version = wgtk.qVersion()
  return qt_version
  
def GetPyDebugBuild():
  """ Return whether code is running in a python with the pydebug build 
  option enabled """
  
  try:
    sys.getobjects
  except AttributeError:
    return False
  
  return True

def LoadDynamicLibraries(log, winghome, name_list, runtime_list=None,
                         optional=()):
  """ Load the named c libraries (.so / .dll) from any runtime in the
  runtime list.  If runtime_list is None, all runtimes under WINGHOME/bin
  are searched """
  
  import ctypes

  if len(name_list) == 0:
    return
  
  if runtime_list is None:
    runtime_list = [n for n in os.listdir(RelativePathJoin(winghome, 'bin'))
                    if n.startswith('runtime-')]
  
  leaf_dir_name = ('bin' if sys.platform == 'win32' else 'lib')
  dynlib_dir_list = [RelativePathJoin(winghome, 'bin', rt_name, leaf_dir_name)
                     for rt_name in runtime_list]

  files_for_dir = {}
  for dynlib_dir in dynlib_dir_list:
    try:
      files_for_dir[dynlib_dir] = os.listdir(dynlib_dir)
    except OSError:
      files_for_dir[dynlib_dir] = []
  
  def find_name(name):
    for dynlib_dir in dynlib_dir_list:
      file_list = files_for_dir[dynlib_dir]
      if name in file_list:
        return RelativePathJoin(dynlib_dir, name)
      
    return None
  
  def load_one(name):
    fullname = find_name(name)
    if fullname is None:
      if name not in optional:
        log('Could not find a file on disk for: {}\n'.format(name))
      return
  
    # XXX 2.7.13 change?
    if isinstance(fullname, unicode):
      fullname = fullname.encode(sys.getfilesystemencoding())
    
    try:
      log("Loading %s\n" % repr(fullname))
      ctypes.cdll.LoadLibrary(fullname)
    except Exception:
      log('Unable to load %s\n' % repr(fullname))
      raise
      
  for name in name_list:
    load_one(name)
    
gSavedStringsForQtConf = None
def InitializePyQt(log, winghome, qt_runtime):
  """ Load PyQt based libraries """
  
  from guiutils import qt_config
  
  # Nested function to setup the in-memory qt.conf; must be called after
  # qt_base but before wgtk is imported
  def register_qt_conf(qt_runtime_dir):
    """ setup the in-memory qt.conf; called after QtCore can be imported
    and before nearly anything else is done. """
    
    import pyside_utils
    from PyQt5 import QtCore
    pyside_utils.QtCore = QtCore

    if sys.platform == 'win32':
      qt_runtime_dir = qt_runtime_dir.replace('\\', '/')
    rc_struct, rc_name, rc_data = pyside_utils._get_qt_conf_resource(
      prefix=qt_runtime_dir, 
      binaries=qt_runtime_dir+"/bin",
      plugins=qt_runtime_dir+"/plugins",
      imports=qt_runtime_dir+"/imports",
      translations=qt_runtime_dir+"/translations")
    # XXX Find a better way to do this; we need to save a ref to the strings
    # passed to qRegisterResourceData
    global gSavedStringsForQtConf
    gSavedStringsForQtConf = (rc_struct, rc_name, rc_data)
    QtCore.qRegisterResourceData(0x01, rc_struct, rc_name, rc_data)
    prefixPath = QtCore.QLibraryInfo.location(QtCore.QLibraryInfo.PrefixPath)
    if sys.platform == 'win32':
      assert os.path.normpath(prefixPath) == os.path.normpath(qt_runtime_dir)
    else:
      assert os.path.samefile(prefixPath, qt_runtime_dir)

  debug_build = GetPyDebugBuild()
  sci_runtime_name = 'runtime-scintillaedit-' + qt_runtime

  qt_runtime_dir = os.path.normpath(winghome + '/bin/runtime-{}'.format(qt_runtime))

  # Code below loads .dll / .so files in 4 steps: before anything qt related,
  # files for QtCore, the rest of qt, and then scintilla.  Any or all of these
  # can be empty (currently all are on OS X)
  runtime_list = ['runtime-{}'.format(qt_runtime)]
  pre_qt_dlls = []
  qt_core_dlls = []
  rest_qt_dlls = []
  sciedit_dynlibs = []
  
  # Names of qt modules ordered by dependency
  qt_core_names = ['Core']
  rest_qt_names = ['Gui', 'Widgets', 'Svg',]
  qt_print_names = ['Network', 'Qml', 'Quick', 'PrintSupport']
  qt_webengine_names = ['WebChannel', 'Positioning',
                        'WebEngineCore', 'WebEngine', 'WebEngineWidgets']
  if sys.platform == 'win32':
    qt_webkit_names = ['Positioning', 'Sensors', 'Sql',
                       'OpenGL', 'Multimedia',  'MultimediaWidgets',
                       'WebChannel', 'WebKit', 'WebKitWidgets']
  else:
    # Multimedia not needed on linux
    qt_webkit_names = ['Positioning', 'Sensors', 'Sql',
                       'OpenGL', 
                       'WebChannel', 'WebKit', 'WebKitWidgets']
    
  rest_qt_names.extend(qt_print_names)
  if qt_config.kUseWebEngine:
    rest_qt_names.extend(qt_webengine_names)
  if qt_config.kUseWebKit:
    rest_qt_names.extend(qt_webkit_names)

  # Fill in the lists depending on the platform
  if sys.platform == 'win32':
    def win32_dll_name(n):
      if debug_build:
        n += 'd'
      return 'Qt5{}.dll'.format(n)
    pre_qt_dlls = ['icudt54.dll', 'icuuc54.dll', 'icuin54.dll']
    if qt_config.kUseWebEngine:
      gl_names = ['GLESv2', 'libEGL']
      pre_qt_dlls.extend(
        ['lib{}'.format(n) + ('d.dll' if debug_build else '.dll')
         for n in gl_names])
    qt_core_dlls = [win32_dll_name(n) for n in qt_core_names]
    rest_qt_dlls = [win32_dll_name(n) for n in rest_qt_names]
    sciedit_dynlibs = ['ScintillaEdit3.dll']
    
  elif sys.platform != 'darwin': # Non OS X unix
    def so_name(n):
      return 'libQt5{}.so.5'.format(n)
    if qt_runtime == 'qt5.5':
      pre_qt_dlls = [
        'libgstreamer-0.10.so.0', 
        'libgstbase-0.10.so.0',
        'libgstapp-0.10.so.0',
        'libgstinterfaces-0.10.so.0',
        'libgstpbutils-0.10.so.0',
        'libgstvideo-0.10.so.0',
        
        # Need our copy of icu -- other versions can cause C++ link errors
        'libicudata.so.54',
        'libicuuc.so.54',
        'libicui18n.so.54',
      ]
      runtime_list += ['runtime-gstreamer0.10', 
                       'runtime-gstreamer-plugins-base0.10']
      # The xcb platform plugin brings in DBus and a .so from the lib dir
      rest_qt_names += ['DBus', 'XcbQpa']
    else:# qt5.8
      pre_qt_dlls = [
        # Need our copy of icu -- other versions can cause C++ link errors
        'libicudata.so',
        'libicuuc.so',
        'libicui18n.so',
      ]
      runtime_list += ['runtime-icu']
      rest_qt_names += ['XcbQpa']

    qt_core_dlls = [so_name(n) for n in qt_core_names]
    rest_qt_dlls = [so_name(n) for n in rest_qt_names]
    sciedit_dynlibs = ['libScintillaEdit.so.3']

  LoadDynamicLibraries(log, winghome, pre_qt_dlls, runtime_list=runtime_list)
  LoadDynamicLibraries(log, winghome, qt_core_dlls, runtime_list=runtime_list)
  register_qt_conf(qt_runtime_dir)
  LoadDynamicLibraries(log, winghome, rest_qt_dlls, runtime_list=runtime_list)
  
  # Import Python wrappers after C libraries are loaded
  from guiutils import qt_base

  # Finally load scintilla
  LoadDynamicLibraries(log, winghome, sciedit_dynlibs,
                       runtime_list=[sci_runtime_name])
  
  def msg_handler(level, ctx, msg, *args):
    print msg
  qt_base.qInstallMessageHandler(msg_handler)
  
  # PyQt5 aborts after printing an exception raised by Python code called from
  # non Python aware C++ code (e.g. virtual methods) unless sys.excepthook
  # is overriden
  orig = sys.excepthook
  def excepthook_wrapper(*args):
    return orig(*args)
  sys.excepthook = excepthook_wrapper

def InitializePalettes():
  """ Set up palettes; needs to be done before FieldDef for palette pref is created
  but after it's safe to load wgtk """

  import config
  from guiutils import wgtk

  _kColorPaletteDirs = [
    os.path.join(config.kWingHome, 'resources', 'palettes'), 
    os.path.join(config.kUserWingDir, 'palettes'), 
  ]
  wgtk.gQtStyleConfiguration._SetPaletteDirs(_kColorPaletteDirs)
  wgtk.gQtStyleConfiguration.SetIconDir(config.GetWingFilename('resources/icons'))
  wgtk.gQtStyleConfiguration.SetHTMLCSSDir(config.GetWingFilename('resources'))
  

def ShowCommandLineDocs(argv):

  import gettext
  _ = gettext.translation('src', fallback = 1).ugettext

  # Set up WINGHOME and bring in config if needed; assumption is that program
  # will exit after this runs
  global kWingHome
  if kWingHome is None:
    kWingHome = FindWingHome(argv)

  sys.path.extend([kWingHome + '/src',
                   kWingHome + '/bin/ide-%s.%s/src.zip' % sys.version_info[:2],
                   kWingHome + '/bin/ide-%s.%s/src' % sys.version_info[:2]])
  # Avoid import config since it gets things from wingutils and
  # the dyn directory support isn't set up
  import vinfo
  import capabilities
  product = vinfo.kProduct
  if capabilities.kAvail101Support:
    execname = 'wing-101'
  elif capabilities.kAvailPersonalSupport:
    execname = 'wing-personal'
  else:
    execname = 'wing'
    
  if sys.platform == 'win32':
    execname += '.exe'
  else:
    if capabilities.kAvail101Support:
      execname += '-'
    execname += vinfo.kVersion[:vinfo.kVersion.rfind('.')]
    
  if sys.platform == 'win32':
    verbose = _("""\
Use console_%s to print a diagnostic log while running.""") % execname
  else:
    verbose = _("""\
--verbose                  Print a diagnostic log while running""")

  docs = _("""Wing %(product)s Command Line Usage:

%(execname)s <args> [<project>.wpr] [file1[:lineno], ...]

Optionally specify a project to open and/or a list of files.  Filenames
may be followed by ':' and a line number to scroll to that line.

Valid arguments are:
  
--prefs-file=<filename>    Set preferences override file.  Any preferences
                           in the file override those in the default user
                           preferences file and any changes made during the
                           session are written to the override file.
                             
--new                      Force creation of a new instance of Wing even if
                           an instance is already running.
--reuse                    Force reusing any existing instance of Wing even
                           if no files were given on the command line.
  
%(verbose)s
--help                     Show this message.
  
See also https://wingware.com/doc/install/command-line-usage  
""") % locals()
  
  sys.__stdout__.write(docs)
  
def Run(argv, squelch_output):
  """ Runs wing.  Note that some modules are intentionally not imported
  until they're needed and / or the environment has been set up correctly. """
  
  import premain
  
  for arg in sys.argv:
    if arg.startswith('--command-reference='):
      return premain._GenerateCommandReference()
  if '--preferences-reference' in sys.argv:
    return premain._GeneratePreferenceReference()
  for arg in sys.argv:
    if arg.startswith('--keymap-reference='):
      return premain._GenerateKeyMapReference()
    
  for arg in sys.argv:
    if arg.startswith('--run-module-main='):
      main_argv = list(sys.argv)
      main_argv.remove(arg)
      return premain.RunModuleMain(arg[len('--run-module-main='):], main_argv[1:])
  
  args = premain.ParseArgv(argv)
  pref_files = premain.FindCmdLinePrefFiles(argv)
  cmdlinefiles = premain.FindCmdLineFiles(argv)
  remote_open = premain.RemoteOpenFiles(args, cmdlinefiles)
  
  # Stop now if existing instance should be used
  if remote_open == 1:
    pass

  # Otherwise, continue with startup
  else:
    import main
    
    # Start execution
    def start():
      main.main(squelch_output, remote_open == 0, pref_files, args, cmdlinefiles)
    if sys.platform != 'win32':
      start()
    else:
      from wingutils import win32misc
      win32misc.InWin32TryExcept(start, '')

def Profile(args, squelch_output):
  """ Run w/ profiler. """

  import config

  print 'Starting profile'
  
  import cProfile, pstats
  import time

  wall_start = time.time()
  
  prof = cProfile.Profile()
  prof.runcall(Run, args, squelch_output)
  
  wall_end = time.time()
  
  print 'Profile completed; wall time in seconds = %.2f' % (wall_end - wall_start)
  
  stats = pstats.Stats(prof)
  stats.strip_dirs()
  stats.sort_stats('cumulative', 'calls')
  stats.print_stats(40)
  stats.sort_stats('time', 'calls')
  stats.print_stats(40)

def Initialize(argv, use_error_log=True):
  """ Set up to run.  Returns whether output should be squelched. """

  global kWingHome

  tmp_errs = []
  def log(s):
    assert tmp_errs != None
    tmp_errs.append(s)

  try:
    # Save original language setting before we muck with it internally
    os.environ['WING_ORIG_LANGUAGE'] = os.environ.get('LANGUAGE', '__undefined__')
  
    # Some initial debug info
    log("=" * 80 + '\n')
    log("STARTING WING with args:\n")
    log(str(sys.argv)+'\n')
    log("sys.executable=%s\n" % sys.executable)
    log("sys.maxint=%s\n" % sys.maxint)
    log("LD_LIBRARY_PATH=%s\n" % os.environ.get('LD_LIBRARY_PATH', "''"))
    log("DYLD_LIBRARY_PATH=%s\n" % os.environ.get('DYLD_LIBRARY_PATH', "''"))
    log("DYLD_FALLBACK_LIBRARY_PATH=%s\n" % os.environ.get('DYLD_FALLBACK_LIBRARY_PATH', "''"))
    log("PATH=%s\n" % os.environ.get('PATH', "''"))
    
    # Find WingHome, note that config.py depends on the kWingHome name    
    kWingHome = FindWingHome(argv)
    
    SetupPythonPath(log)
    
    log("PYTHONPATH=%s\n" % str(sys.path))

    InitializeDynamicPaths(log)
    assert PATCHES_INSTALLED
    
    # Code in wingbase uses wingversion to inspect what patches are installed
    from wingbase import wingversion
    import minimalconfig
    wingversion.kPatches = minimalconfig.gPatchPairList

    InitializeSteam(log, kWingHome)

    InitializeConfig()

    qt_version = InitializeQt(log, kWingHome)
    log("Qt version=%s\n" % qt_version)
    
    InitializeDebugServer()
  
    squelch_output = GetSquelchOutputFlag()
    if use_error_log:
      InitializeStdoutLogging(squelch_output)
  
    InitializeFaultHandler(log, minimalconfig.kUserWingDir)
    
    # Print any accumulated errors to the log
    for err in tmp_errs:
      sys.stderr.write(err)
    tmp_errs = None
  
    return squelch_output

  except:
    # If strings in tmp_errs, send them to stderr
    if tmp_errs:
      for err in tmp_errs:
        sys.stderr.write(err)
    raise
  
def Main(argv):
  """ Setup & run wing. """

  # Work around for Python codec bug on Windows (only affects Wing running as console app)
  if sys.platform == 'win32':
    import codecs
    codecs.register(lambda name: codecs.lookup('utf-8') if name == 'cp65001' else None)
    
  if any(arg.startswith('--help') for arg in argv):
    ShowCommandLineDocs(argv)
    return

  squelch_output = Initialize(argv)

  # Set up I18N 
  import gettext
  _ = gettext.translation('src', fallback = 1).ugettext

  profile = GetProfileFlag()
  debug = GetDebugFlag()
  if profile:
    Profile(argv, squelch_output)
    return
  
  if debug:
    Run(argv, squelch_output)
  
  else:
  
    try:
      Run(argv, squelch_output)
    except:
      print _("Fatal startup error:  Exiting Wing")
      raise

  # XXXQT Avoid shutdown bugs when not debugging
  if not __debug__:
    if sys.platform == 'win32':
      # _exit() on win32 still seems to call the atexit functions or
      # some sort of destructors in shiboken
      from wingbase import win32
      win32.TerminateProcess(win32.GetCurrentProcess(), 0)
    os._exit(0)
  
if __name__ == '__main__':
  try:
    Main(list(sys.argv))
  finally:
    # Make sure any pending log entries are written to log files
    from wingutils import reflect
    reflect._DoWrite()
    
    
